/*	$Id: audio.c,v 1.25 2002/06/03 15:46:50 nonaka Exp $	*/

/*
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *      This product includes software developed by NONAKA Kimihiro.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "compiler.h"

#include "audio.h"
#include "ini.h"

static BOOL noaudio_init(DWORD, DWORD);
static BOOL noaudio_term(void);
static void noaudio_play(void);
static void noaudio_stop(void);

int audio_fd = -1;
audiodrv_t audiodrv = {
	noaudio_init,
	noaudio_term,
	noaudio_play,
	noaudio_stop
};

static DWORD opna_rate = 22050;
static DWORD opna_frame = 2048;

int audio_nextbuf;
char *audio_event;
char *audio_buffer[NSOUNDBUFFER];

char audiodev[MAX_PATH];

static void audio_send(void);

static BOOL buffer_init(DWORD, DWORD);
static void buffer_cleanup(void);

/*--
 * np2osdep
 */
BOOL
np2os_waveplay(int num, int lpflg)
{

	UNUSED(num);
	UNUSED(lpflg);

	return FAILURE;
}

void
np2os_wavestop(int num)
{

	UNUSED(num);

	/* Nothing to do */
}

BYTE
audio_drv2num(char *cfgstr)
{

	if (strcasecmp(cfgstr, "NetBSD") == 0)
		return SNDDRV_NETBSD;
	else if (strcasecmp(cfgstr, "OSS") == 0)
		return SNDDRV_OSS;
	else if (strcasecmp(cfgstr, "ESD") == 0)
		return SNDDRV_ESD;
	else if (strcasecmp(cfgstr, "SDL") == 0)
		return SNDDRV_SDL;
	else if (strcasecmp(cfgstr, "") == 0)
		return SNDDRV_DRVMAX;
	return SNDDRV_NODRV;
}

char *
audio_num2drv(BYTE num)
{

	switch (num) {
	case SNDDRV_NETBSD:
		return "NetBSD";
	case SNDDRV_OSS:
		return "OSS";
	case SNDDRV_ESD:
		return "ESD";
	case SNDDRV_SDL:
		return "SDL";
	case SNDDRV_DRVMAX:
		return "";
	}
	return "nosound";
}

DWORD
calc_blocksize(DWORD size)
{
	DWORD s = size;

	if (size & (size - 1))
		for (s = 32; s < size; s <<= 1)
			continue;
	return s;
}

int
calc_fragment(DWORD size)
{
	DWORD f;

	for (f = 0; size > (DWORD)(1 << f); f++)
		continue;
	return f;
}

void
audio_setup(void)
{

	if (np2cfg.snddrv <= SNDDRV_DRVMAX) {
		switch (np2cfg.snddrv) {
#if defined(USE_NETBSDAUDIO)
		case SNDDRV_NETBSD:
			netbsdaudio_setup();
			break;
#endif
#if defined(USE_OSSAUDIO)
		case SNDDRV_OSS:
			ossaudio_setup();
			break;
#endif
#if defined(USE_ESDAUDIO)
		case SNDDRV_ESD:
			esdaudio_setup();
			break;
#endif
#if defined(USE_SDLAUDIO)
		case SNDDRV_SDL:
			sdlaudio_setup();
			break;
#endif
		}
	} else {
#if defined(USE_NETBSDAUDIO)
		if (netbsdaudio_setup() == SUCCESS) {
			np2cfg.snddrv = SNDDRV_NETBSD;
			INIT_RENEWAL;
		} else
#endif
#if defined(USE_OSSAUDIO)
		if (ossaudio_setup() == SUCCESS) {
			np2cfg.snddrv = SNDDRV_OSS;
			INIT_RENEWAL;
		} else
#endif
#if defined(USE_ESDAUDIO)
		if (esdaudio_setup() == SUCCESS) {
			np2cfg.snddrv = SNDDRV_ESD;
			INIT_RENEWAL;
		} else
#endif
#if defined(USE_SDLAUDIO)
		if (sdlaudio_setup() == SUCCESS) {
			np2cfg.snddrv = SNDDRV_SDL;
			INIT_RENEWAL;
		} else
#endif
		{
			np2cfg.snddrv = SNDDRV_NODRV;
			INIT_RENEWAL;
		}
	}
}

BOOL
audio_init(DWORD rate, DWORD bufmsec)
{

	if (rate == 0)
		return SUCCESS;

	if (rate < AVE(11025, 22050))
		opna_rate = 11025;
	else if (rate < AVE(22050, 44100))
		opna_rate = 22050;
	else
		opna_rate = 44100;

	if (bufmsec < 50)
		bufmsec = 50;
	else if (bufmsec > 1000)
		bufmsec = 1000;

	opna_frame = calc_blocksize(opna_rate * bufmsec / 1000 / 2);

	audio_nextbuf = 0;
	audio_event = NULL;

	if ((*audiodrv.init)(opna_rate, opna_frame) != SUCCESS) {
		np2cfg.snddrv = SNDDRV_NODRV;
		INIT_RENEWAL;
	}
	buffer_init(opna_rate, opna_frame);

	np2sound_init(opna_rate, audio_send);

	return SUCCESS;
}

BOOL
audio_term(void)
{

	(*audiodrv.term)();
	buffer_cleanup();
	audio_fd = -1;

	return SUCCESS;
}

void
audio_play(void)
{

	if (audio_fd < 0)
		return;

	(*audiodrv.play)();
}

void
audio_stop(void)
{

	if (audio_fd < 0)
		return;

	(*audiodrv.stop)();
}

/*
 * Not sound support
 */
static BOOL
noaudio_init(DWORD rate, DWORD bufmsec)
{

	UNUSED(rate);
	UNUSED(bufmsec);

	return SUCCESS;
}

static BOOL
noaudio_term(void)
{

	return SUCCESS;
}

static void
noaudio_play(void)
{

	/* Nothing to do */
}

static void
noaudio_stop(void)
{

	/* Nothing to do */
}

/*
 * common audio portion
 */
static void
audio_send(void)
{
	NP2SNDBUF_T sndbuf;
	long *p;
	short *q;
	long smp;
	int remain;

	if (audio_fd < 0)
		return;

	if (audio_event != NULL) {
		q = (short *)audio_event;
		audio_event = NULL;

		np2sound_getsounddata(opna_frame, &sndbuf);

		if (sndbuf.len1 != 0) {
			p = sndbuf.buf1;
			remain = sndbuf.len1 * 2;

			while (remain-- > 0) {
				smp = *p++;
				if (smp > 32767)
					smp = 32767;
				else if (smp < -32767)
					smp = -32767;
				*q++ = (short)smp;
			}
		}
		if (sndbuf.len2 != 0) {
			p = sndbuf.buf2;
			remain = sndbuf.len2 * 2;

			while (remain-- > 0) {
				smp = *p++;
				if (smp > 32767)
					smp = 32767;
				else if (smp < -32767)
					smp = -32767;
				*q++ = (short)smp;
			}
		}
	}
}

static BOOL
buffer_init(DWORD rate, DWORD samples)
{
	int i;

	UNUSED(rate);

	for (i = 0; i < NSOUNDBUFFER; i++) {
		audio_buffer[i] = (char *)calloc(samples * 4, 1);
		if (audio_buffer[i] == NULL) {
			fprintf(stderr, "buffer_init: can't alloc memory\n");
			return FAILURE;
		}
	}

	return SUCCESS;
}

static void
buffer_cleanup(void)
{
	int i;

	for (i = 0; i < NSOUNDBUFFER; i++) {
		if (audio_buffer[i] != NULL)
			free(audio_buffer[i]);
	}
}

#if defined(USE_NETBSDAUDIO) || defined(USE_OSSAUDIO) || defined(USE_ESDAUDIO)
static pthread_t ptid;
static int is_proc;

static void *buffer_play(void *);
static void buffer_write(int, void *, size_t);

static void
audio_callback(void *userdata, unsigned char *stream, int len)
{

	UNUSED(userdata);
	UNUSED(len);

	if (audio_event != NULL)
		memset(audio_event, 0, opna_frame * 4);

	memcpy(stream, audio_buffer[audio_nextbuf], opna_frame * 4);

	if (++audio_nextbuf >= NSOUNDBUFFER)
		audio_nextbuf = 0;
	audio_event = audio_buffer[audio_nextbuf];
}

static void
buffer_write(int fd, void *buf, size_t len)
{
	ssize_t r;
	size_t s = 0;

	for (;;) {
		r = write(fd, buf + s, len - s);
		if (r >= 0) {
			s += r;
			if (len <= s)
				break;
		}
	}
}

static void *
buffer_play(void *arg)
{
	unsigned char *buf;
	size_t len;

	UNUSED(arg);

	len = opna_frame * 4;
	buf = (unsigned char *)calloc(len, 1);
	if (buf == NULL) {
		fprintf(stderr, "buffer_play: can't alloc memory\n");
		return NULL;
	}

	is_proc = 1;
	while (is_proc) {
		audio_callback(NULL, buf, len);
		buffer_write(audio_fd, buf, len);
	}
	free(buf);
	is_proc = 0;

	return NULL;
}

void
audio_generic_play(void)
{

	if (pthread_create(&ptid, NULL, buffer_play, NULL) != 0) {
		fprintf(stderr, "audio_play: can't create thread.\n");
	}
}

void
audio_generic_stop(void)
{

	is_proc = 0;
}

#endif	/* USE_NETBSDAUDIO || USE_OSSAUDIO || USE_ESDAUDIO */
