/*	$Id: midi.c,v 1.3 2002/07/09 16:01:58 nonaka Exp $	*/

#include "compiler.h"

#include <fcntl.h>
#include <sys/time.h>

#include "np2.h"
#include "midi.h"

#define MIDI_RESETWAIT	80000

#define	MIDICTRL_READY		0
#define	MIDICTRL_2BYTES		1
#define	MIDICTRL_3BYTES		2
#define	MIDICTRL_EXCLUSIVE	3
#define	MIDICTRL_TIMECODE	4
#define MIDICTRL_SYSTEM		5

#define	MIDI_EXCLUSIVE		0xf0
#define MIDI_TIMECODE		0xf1
#define MIDI_SONGPOS		0xf2
#define MIDI_SONGSELECT		0xf3
#define	MIDI_TUNEREQUEST	0xf6
#define	MIDI_EOX		0xf7
#define	MIDI_TIMING		0xf8
#define MIDI_START		0xfa
#define MIDI_CONTINUE		0xfb
#define	MIDI_STOP		0xfc
#define	MIDI_ACTIVESENSE	0xfe
#define	MIDI_SYSTEMRESET	0xff


char MIDIDEV[MAX_PATH] = "";
int MIDI_WAITTIME = 0;


enum {
	MIDI_MT32 = 0,
	MIDI_CM32L,
	MIDI_CM64,
	MIDI_CM300,
	MIDI_CM500LA,
	MIDI_CM500GS,
	MIDI_SC55,
	MIDI_SC88,
	MIDI_LA,
	MIDI_GM,
	MIDI_GS,
	MIDI_XG,
	MIDI_DEVMAX,
	MIDINUM_DEFAULT = 0xff
};

static const char *midimdl_name[] = {
	"MT-32",	"CM-32L",	"CM-64",
	"CM-300",	"CM-500(LA)",	"CM-500(GS)",
	"SC-55",	"SC-88",	"LA",
	"GM",		"GS",		"XG"
};

static const BYTE EXCV_MTRESET[] = {
	0xfe, 0xfe, 0xfe
};

static const BYTE EXCV_GMRESET[] = {
	0xf0, 0x7e, 0x7f, 0x09, 0x01, 0xf7
};

static const BYTE EXCV_GSRESET[] = {
	0xf0, 0x41, 0x10, 0x42, 0x12, 0x40, 0x00, 0x7f, 0x00, 0x41, 0xf7
};

static const BYTE EXCV_XGRESET[] = {
	0xf0, 0x43, 0x10, 0x4C, 0x00, 0x00, 0x7E, 0x00, 0xf7
};

static int midi_fd = -1;
static BYTE midi_buf[MIDIBUFFERS];
static int midibuf_cur = 0;
static int midi_syscnt = 0;
static BYTE midi_ctrl = MIDICTRL_READY;
static BYTE midi_lastcmd = 0x80;
static struct timeval midi_nextstart;

static void
MIDI_write(const BYTE *midicmd, int cnt)
{
	struct timeval ct;
	int rv;
	int i;

	if (midi_fd < 0)
		return;

	if (MIDI_WAITTIME >= 0) {
		for (;;) {
			gettimeofday(&ct, NULL);
			if (ct.tv_sec > midi_nextstart.tv_sec)
				break;

			if (ct.tv_sec == midi_nextstart.tv_sec)
				if (ct.tv_usec >= midi_nextstart.tv_usec)
					break;
		}
	}

	for (i = 0; i < cnt; i++) {
		do {
			rv = write(midi_fd, midicmd + i, 1);
		} while (rv != 1);
	}

	if (MIDI_WAITTIME < 0)
		return;

	gettimeofday(&midi_nextstart, NULL);
	midi_nextstart.tv_usec += MIDI_WAITTIME * cnt;

	if ((memcmp(midicmd, EXCV_GMRESET, sizeof(EXCV_GMRESET)) == 0)
	    || (memcmp(midicmd, EXCV_GSRESET, sizeof(EXCV_GSRESET)) == 0)
	    || (memcmp(midicmd, EXCV_XGRESET, sizeof(EXCV_XGRESET)) == 0)
	    || (memcmp(midicmd, EXCV_MTRESET, sizeof(EXCV_MTRESET)) == 0)) {
		midi_nextstart.tv_usec += MIDI_RESETWAIT;
	}
	while (midi_nextstart.tv_usec >= 1000000) {
		midi_nextstart.tv_usec -= 1000000;
		midi_nextstart.tv_sec++;
	}
}

void
MIDI_Init(void)
{

	if (midi_fd >= 0)
		return;

	if (MIDIDEV[0] == 0)
		return;

	midi_fd = open(MIDIDEV, O_WRONLY | O_NONBLOCK);
	if (midi_fd < 0) {
		perror("midi_init: open");
		midi_fd = -1;
		return;
	}

	midi_ctrl = MIDICTRL_READY;
	midi_lastcmd = 0x80;
	midibuf_cur = 0;
	gettimeofday(&midi_nextstart, NULL);
}

static int
MIDI_is_ready(void)
{

	if (midi_fd < 0)
		return 0;
	return 1;
}

void
MIDI_AllToneOff(void)
{
	BYTE midicmd[3];
	int i;

	midicmd[1] = 0x7b;
	midicmd[2] = 0x00;
	for (i = 0; i < 16; i++){
		midicmd[0] = 0xb0 + i;
		MIDI_write(midicmd, 3);
	}
}

void
MIDI_Reset(void)
{

	switch (np2cfg.MIDIMODULE) {
	case MIDI_GM:
		MIDI_write(EXCV_GMRESET, sizeof(EXCV_GMRESET));
		break;

	case MIDI_CM300:
	case MIDI_CM500GS:
	case MIDI_SC55:
	case MIDI_SC88:
	case MIDI_GS:
		MIDI_write(EXCV_GSRESET, sizeof(EXCV_GSRESET));
		break;

	case MIDI_XG:
		MIDI_write(EXCV_XGRESET, sizeof(EXCV_XGRESET));
		break;

	case MIDI_MT32:
	case MIDI_CM32L:
	case MIDI_CM64:
	case MIDI_CM500LA:
	case MIDI_LA:
		MIDI_write(EXCV_MTRESET, sizeof(EXCV_MTRESET));
		break;
	}
	MIDI_AllToneOff();
}

void
MIDI_Cleanup(void)
{

	MIDI_AllToneOff();
	close(midi_fd);
	midi_fd = -1;
}

void
np2os_midireset(void)
{

	MIDI_Reset();
}

int
np2os_ismidiready(void)
{

	return MIDI_is_ready();
}

BYTE
MIDI_module2num(char *name)
{
	int i;

	for (i = 0; i < NELEMENTS(midimdl_name); i++)
		if (strcasecmp(name, midimdl_name[i]) == 0)
			return i;
	return MIDINUM_DEFAULT;
}

char *
MIDI_num2module(BYTE midinum)
{

	if (midinum < MIDI_DEVMAX)
		return (char *)midimdl_name[midinum];
	return "unknown";
}

int
np2os_midimessage(BYTE mes)
{

	if (midi_fd < 0)
		return 0;

	switch (mes) {				// бϤߤ
	case MIDI_TIMING:
	case MIDI_START:
	case MIDI_CONTINUE:
	case MIDI_STOP:
	case MIDI_ACTIVESENSE:
		return 1;

	case MIDI_SYSTEMRESET:			// ꡼
		return 1;
	}

	if (midi_ctrl == MIDICTRL_READY) {	// 
		if (mes & 0x80) {
			// status
			midibuf_cur = 0;

			switch(mes & 0xf0) {
			case 0xc0:
			case 0xd0:
				midi_ctrl = MIDICTRL_2BYTES;
				break;

			case 0x80:
			case 0x90:
			case 0xa0:
			case 0xb0:
			case 0xe0:
				midi_ctrl = MIDICTRL_3BYTES;
				midi_lastcmd = mes; // Ԥʤʤ
				break;

			default:
				switch(mes) {
				case MIDI_EXCLUSIVE:
					midi_ctrl = MIDICTRL_EXCLUSIVE;
					break;

				case MIDI_TIMECODE:
					midi_ctrl = MIDICTRL_TIMECODE;
					break;

				case MIDI_SONGPOS:
					midi_ctrl = MIDICTRL_SYSTEM;
					midi_syscnt = 3;
					break;

				case MIDI_SONGSELECT:
					midi_ctrl = MIDICTRL_SYSTEM;
					midi_syscnt = 2;
					break;

				case MIDI_TUNEREQUEST:
					midi_ctrl = MIDICTRL_SYSTEM;
					midi_syscnt = 1;
					break;

//				case MIDI_EOX:
				default:
					return 1;
				}
				break;
			}
		} else {		// Key-onΤߤʵ˺줿
			// running status
			midi_buf[0] = midi_lastcmd;
			midibuf_cur = 1;
			midi_ctrl = MIDICTRL_3BYTES;
		}
	}

	midi_buf[midibuf_cur] = mes;
	midibuf_cur++;

	switch(midi_ctrl) {
	case MIDICTRL_2BYTES:
		if (midibuf_cur >= 2) {
			MIDI_write(midi_buf, 2);
			midi_ctrl = MIDICTRL_READY;
			return 2;
		}
		break;

	case MIDICTRL_3BYTES:
		if (midibuf_cur >= 3) {
			MIDI_write(midi_buf, 3);
			midi_ctrl = MIDICTRL_READY;
			return 3;
		}
		break;

	case MIDICTRL_EXCLUSIVE:
		if (mes == MIDI_EOX) {
			MIDI_write(midi_buf, midibuf_cur);
			midi_ctrl = MIDICTRL_READY;
			return midibuf_cur;
		} else if (midibuf_cur >= MIDIBUFFERS) { // Сդ
			midi_ctrl = MIDICTRL_READY;
		}
		break;

	case MIDICTRL_TIMECODE:
		if (midibuf_cur >= 2) {
			if ((mes == 0x7e) || (mes == 0x7f)) {
				// exclusiveƱǤȦ
				midi_ctrl = MIDICTRL_EXCLUSIVE;
			} else {
				midi_ctrl = MIDICTRL_READY;
				return 2;
			}
		}
		break;

	case MIDICTRL_SYSTEM:
		if (midibuf_cur >= midi_syscnt) {
			midi_ctrl = MIDICTRL_READY;
			return midi_syscnt;
		}
		break;
	}

	return 0;
}
